@@ -43,14 +43,14 @@ from typing import Any, Optional, Tuple, List, Set, Dict, Union, Callable, cast
 
 from DHParser.configuration import get_config_value
 from DHParser.preprocess import PreprocessorFunc
-from DHParser.syntaxtree import Node, RootNode, EMPTY_PTYPE, TreeContext
+from DHParser.nodetree import Node, RootNode, EMPTY_PTYPE, TreeContext
 from DHParser.transform import TransformerCallable
 from DHParser.parse import Grammar
 from DHParser.preprocess import gen_neutral_srcmap_func
 from DHParser.error import is_error, is_fatal, Error, FATAL, \
     TREE_PROCESSING_CRASH, COMPILER_CRASH, AST_TRANSFORM_CRASH, has_errors
 from DHParser.log import log_parsing_history, log_ST, is_logging
-from DHParser.toolkit import load_if_file, is_filename
+from DHParser.toolkit import load_if_file, is_filename, re
 
 
 __all__ = ('CompilerError',
@@ -77,28 +77,37 @@ class CompilerError(Exception):
 
 def visitor_name(node_name: str) -> str:
     """
-    Returns the visitor_method name for `node_name`, e.g.::
+    Returns the visitor_method name for `name`, e.g.::
 
-    >>> visitor_name('expression')
-    'on_expression'
+        >>> visitor_name('expression')
+        'on_expression'
+        >>> visitor_name('!--')
+        'on_212d2d'
     """
-    # assert re.match(r'\w+$', node_name)
-    return 'on_' + node_name
+    if re.match(r'\w+$', node_name):
+        return 'on_' + node_name
+    else:
+        letters = []
+        for ch in node_name:
+            if not re.match(r'\w', ch):
+                ch = hex(ord(ch))[2:]
+            letters.append(ch)
+        return 'on_' + ''.join(letters)
 
 
 def attr_visitor_name(attr_name: str) -> str:
     """
     Returns the visitor_method name for `attr_name`, e.g.::
 
-    >>> attr_visitor_name('class')
-    'attr_class'
+        >>> attr_visitor_name('class')
+        'attr_class'
     """
-    # assert re.match(r'\w+$', node_name)
     return 'attr_' + attr_name
 
 
 
 ROOTNODE_PLACEHOLDER = RootNode()
+CompilerFunc = Callable[[Node], Any]
 
 
 class Compiler:
@@ -121,9 +130,6 @@ class Compiler:
     method which will pick the right `on_XXX`-method. It is not
     recommended to call the `on_XXX`-methods directly.
 
-    :ivar source: The source text of the AST to be compiled. This needs to be
-                assigned by the user of the Compiler object - as is done
-                by function `compile_source()`
     :ivar context:  A list of parent nodes that ends with the currently
                 compiled node.
     :ivar tree: The root of the abstract syntax tree.
@@ -153,7 +159,6 @@ class Compiler:
         self.reset()
 
     def reset(self):
-        self.source = ''  # type: str
         self.tree = ROOTNODE_PLACEHOLDER   # type: RootNode
         self.context = []  # type: TreeContext
         self._None_check = True  # type: bool
@@ -161,6 +166,7 @@ class Compiler:
         self._debug = get_config_value('debug_compiler')  # type: bool
         self._debug_already_compiled = set()              # type: Set[Node]
         self.finalizers = []  # type: List[Tuple[Callable, Tuple]]
+        self.method_dict = {}  # type: Dict[str, CompilerFunc]
 
     def prepare(self, root: Node) -> None:
         """
@@ -184,12 +190,27 @@ class Compiler:
         the compiler to determine the format of the returned data.
         (This very much depends on the kind and purpose of the
         implemented compiler.)
+
+        The ``__call__``-method is also responsible for initializations
+        required before the compilation and the finalization the compilation
+        has been finished by taking the following steps::
+
+            1. reset all variables and initalize ``self.tree`` with ``root``
+            2. call the :py:meth:`Compiler.prepare()`-method.
+            3. compile the syntax-tree originating in ``root`` by calling
+               :py:meth:`Compiler.compile()` on the root-node.
+            4. call all finalizers in the ``self.finalizers``-list.
+            5. call the :py:meth:`Compiler.finalize()`-method.
+
+        :param root: The root-node of the syntax tree to be compiled. ``root``
+            does not need to be of type :py:class:`~nodetree.RootNode` in order
+            to allow compiling parts of a syntax tree.
+        :returns: The resulting object of the compilation.
         """
         if self._dirty_flag:
             self.reset()
         self._dirty_flag = True
-        self.tree = root
-        # self.source = source  # type: str
+        self.tree = cast(RootNode, root) if isinstance(root, RootNode) else RootNode(root)
         self.prepare(root)
         result = self.compile(root)
         while self.finalizers:
@@ -198,35 +219,9 @@ class Compiler:
         result = self.finalize(result)
         return result
 
-    def fallback_compiler(self, node: Node, block_attribute_visitors: bool=False) -> Any:
-        """This is a generic compiler function which will be called on
-        all those node types for which no compiler method `on_XXX` has
-        been defined."""
-        replacements = {}  # type: Dict[Node, Node]
-        if node.children:
-            for child in node.children:
-                nd = self.compile(child)
-                if id(nd) != id(child):
-                    replacements[id(child)] = nd
-                if nd is not None and not isinstance(nd, Node):
-                    tn = node.tag_name
-                    raise TypeError(
-                        'Fallback compiler for Node `%s` received a value of type '
-                        '`%s` from child `%s` instead of the required return type `Node`. '
-                        'Override `DHParser.compile.Compiler.fallback_compiler()` or add '
-                        'method `on_%s(self, node)` in class `%s` to avoid this error!'
-                        % (tn, str(type(nd)), child.tag_name, tn, self.__class__.__name__))
-            if replacements:
-                # replace Nodes the identity of which has been changed during transformation
-                # and drop any returned None-results
-                result = []
-                for child in node.children:
-                    nd = replacements.get(id(child), child)
-                    if nd is not None and nd.tag_name != EMPTY_PTYPE:
-                        result.append(nd)
-                node.result = tuple(result)
-        if self.has_attribute_visitors and not block_attribute_visitors and node.has_attr():
-            for attribute, value in node.attr.items():
+    def visit_attributes(self, node) -> Node:
+        if node.has_attr():
+            for attribute, value in list(node.attr.items()):
                 try:
                     attribute_visitor = self.__getattribute__(attr_visitor_name(attribute))
                     node = attribute_visitor(node, value) or node
@@ -234,6 +229,49 @@ class Compiler:
                     pass
         return node
 
+    def fallback_compiler(self, node: Node, block_attribute_visitors: bool=False) -> Any:
+        """This is a generic compiler function which will be called on
+        all those node types for which no compiler method `on_XXX` has
+        been defined."""
+        replacements = {}  # type: Dict[int, Node]
+        if node.children:
+            for child in node.children:
+                nd = self.compile(child)
+                if id(nd) != id(child):
+                    replacements[id(child)] = nd
+                if nd is not None and not isinstance(nd, Node):
+                    tn = node.name
+                    raise TypeError(
+                        'Fallback compiler for Node `%s` received a value of type '
+                        '`%s` from child `%s` instead of the required return type `Node`. '
+                        'Override `DHParser.compile.Compiler.fallback_compiler()` or add '
+                        'method `on_%s(self, node)` in class `%s` to avoid this error!'
+                        % (tn, str(type(nd)), child.name, tn, self.__class__.__name__))
+            if replacements:
+                # replace Nodes the identity of which has been changed during transformation
+                # and drop any returned None-results
+                result = []
+                for child in node.children:
+                    nd = replacements.get(id(child), child)
+                    if nd is not None and nd.name != EMPTY_PTYPE:
+                        result.append(nd)
+                node.result = tuple(result)
+        if self.has_attribute_visitors and not block_attribute_visitors:
+            node = self.visit_attributes(node)
+        return node
+
+    def get_compiler(self, node_name: str) -> CompilerFunc:
+        try:
+            method = self.method_dict[node_name]
+        except KeyError:
+            method_name = visitor_name(node_name)
+            try:
+                method = self.__getattribute__(method_name)
+            except AttributeError:
+                method = self.fallback_compiler
+            self.method_dict[node_name] = method
+        return method
+
     def compile(self, node: Node) -> Any:
         """
         Calls the compilation method for the given node and returns the
@@ -251,21 +289,17 @@ class Compiler:
             assert node not in self._debug_already_compiled
             self._debug_already_compiled.add(node)
 
-        elem = node.tag_name
+        elem = node.name
         if elem[:1] == ':':
             elem = elem[1:] + '__'
-        try:
-            compiler = self.__getattribute__(visitor_name(elem))
-            # print(self.__class__.__name__, elem, str(node)[:80])
-        except AttributeError:
-            compiler = self.fallback_compiler
+        compiler = self.get_compiler(elem)
         self.context.append(node)
         result = compiler(node)
         self.context.pop()
         if result is None and self._None_check:
             raise CompilerError(
                 ('Method on_%s returned `None` instead of a valid compilation '
-                 'result! It is recommended to use `syntaxtree.EMPTY_NODE as a '
+                 'result! It is recommended to use `nodetree.EMPTY_NODE` as a '
                  'void value. This Error can be turn off by adding '
                  '`self._None_check = False` to the reset()-Method of your'
                  'compiler class, in case on_%s actually SHOULD be allowed to '
@@ -333,7 +367,8 @@ def process_tree(tp: CompilerCallable, tree: RootNode) -> Any:
             try:
                 result = tp(tree)
             except Exception as e:
-                node = tp.context[-1] if tp.context else tree
+                # node = tp.context[-1] if hasattr(tp, 'context') and tp.context else tree
+                node = getattr(tp, 'context', [tree])[-1]
                 st = traceback.format_list(traceback.extract_tb(e.__traceback__))
                 trace = ''.join(st)  # filter_stacktrace(st)
                 tree.new_error(
@@ -404,7 +439,7 @@ def run_pipeline(junctions: Set[Junction],
 CompilationResult = namedtuple('CompilationResult',
     ['result',      # type: Optional[Any]
      'messages',    # type: List[Error]
-     'AST'],        # type: RootNode
+     'AST'],        # type: Optional[RootNode]
     module=__name__)
 
 
@@ -446,10 +481,11 @@ def compile_source(source: str,
            or `None` otherwise.
     """
     ast = None  # type: Optional[Node]
-    original_text = load_if_file(source)  # type: str
-    source_name = source if is_filename(source) else 'source'
-    if isinstance(compiler, Compiler):
-        compiler.source = original_text
+    try:
+        original_text = load_if_file(source)  # type: str
+    except (FileNotFoundError, IOError) as e:
+        return CompilationResult(None, [Error(str(e), 0, COMPILER_CRASH)], None)
+    source_name = source if is_filename(source) else ''
     log_file_name = logfile_basename(source, compiler) if is_logging() else ''  # type: str
     if not hasattr(parser, 'free_char_parsefunc__') or parser.history_tracking__:
         # log only for custom parser/transformer/compilers
@@ -514,28 +550,6 @@ def compile_source(source: str,
 
             result = process_tree(compiler, syntax_tree)
 
-            # if is_error(syntax_tree.error_flag):
-            #     # assume Python crashes are merely a consequence of earlier
-            #     # errors, so let's catch them
-            #     try:
-            #         result = compiler(syntax_tree)
-            #     except Exception as e:
-            #         # raise e
-            #         node = syntax_tree  # type: Node
-            #         if isinstance(compiler, Compiler) and compiler.context:
-            #             node = compiler.context[-1]
-            #         st = traceback.format_list(traceback.extract_tb(e.__traceback__))
-            #         trace = ''.join(filter_stacktrace(st))
-            #         syntax_tree.new_error(
-            #             node, "Compilation failed, most likely, due to errors earlier "
-            #                   "in the processing pipeline. Crash Message: %s: %s\n%s"
-            #                   % (e.__class__.__name__, str(e), trace),
-            #             COMPILER_CRASH)
-            # else:
-            #     # assume Python crashes are programming mistakes, so let
-            #     # the exceptions through
-            #     result = compiler(syntax_tree)
-
     messages = syntax_tree.errors_sorted  # type: List[Error]
     # Obsolete, because RootNode adjusts error locations whenever an error is added:
     # adjust_error_locations(messages, original_text, source_mapping)
@@ -544,4 +558,5 @@ def compile_source(source: str,
 
 # TODO: Verify compiler against grammar,
 #       i.e. make sure that for all on_X()-methods, `X` is the name of a parser
+#       Does that make sense? Tag names could change during AST-Transformation!
 # TODO: AST validation against an ASDL-Specification
