@@ -53,16 +53,15 @@ Example::
 
 import collections
 import contextlib
-import copy
 import html
 import os
-from typing import List, Tuple, Union, Optional
+from typing import List, Tuple, Union, Optional, Sequence
 
 from DHParser.configuration import access_presets, finalize_presets, get_config_value, \
     set_config_value, get_preset_value, set_preset_value
 from DHParser.error import Error
 from DHParser.stringview import StringView
-from DHParser.syntaxtree import Node, FrozenNode, ZOMBIE_TAG, EMPTY_PTYPE
+from DHParser.nodetree import Node, FrozenNode, ZOMBIE_TAG, EMPTY_PTYPE
 from DHParser.toolkit import escape_ctrl_chars, abbreviate_middle, cython
 
 __all__ = ('CallItem',
@@ -77,12 +76,13 @@ __all__ = ('CallItem',
            'clear_logs',
            'NONE_TAG',
            'NONE_NODE',
+           'callstack_as_str',
            'HistoryRecord',
            'log_ST',
            'log_parsing_history')
 
 
-CallItem = Tuple[str, int]      # call stack item: (tag_name, location)
+CallItem = Tuple[str, int]      # call stack item: (name, location)
 
 
 #######################################################################
@@ -273,6 +273,24 @@ NONE_TAG = ":None"
 NONE_NODE = NoneNode(NONE_TAG, '')
 
 
+def callstack_as_str(callstack: Sequence[CallItem], depth=-1) -> str:
+    """Returns a string representation of the callstack!"""
+    # return "->".join(name for name, _ in self.call_stack)
+    short_stack = []
+    anonymous_tail = True
+    for node_name, _ in reversed(callstack):
+        if node_name[:1] == ':':
+            if anonymous_tail and node_name != ":Forward":
+                short_stack.append(node_name)
+        else:
+            short_stack.append(node_name)
+            anonymous_tail = False
+    if depth <= 0:  depth = len(short_stack)
+    s = "->".join(reversed(short_stack[:depth]))
+    omitted = len(short_stack) - depth
+    return f'[{omitted}]->' + s if omitted > 0 else s
+
+
 class HistoryRecord:
     """
     Stores debugging information about one completed step in the
@@ -370,7 +388,8 @@ class HistoryRecord:
         if status in (self.MATCH, self.DROP):
             n = max(40 - len(excerpt), 0)
             dots = '...' if len(self.text) > n else ''
-            excerpt = excerpt + '<span class="unmatched">' + self.text[:n] + dots + '</span>'
+            excerpt = ''.join([excerpt, '<span class="unmatched">',
+                               html.escape(str(self.text[:n])), dots, '</span>'])
             i = stack.rfind('-&gt;')
             ch = stack[i + 12:i + 13]
             while not ch.isidentifier() and i >= 0:
@@ -380,16 +399,16 @@ class HistoryRecord:
                 i += 12
                 k = stack.find('<', i)
                 if k < 0:
-                    stack = stack[:i] + '<span class="matchstack">' + stack[i:]
+                    stack = ''.join([stack[:i], '<span class="matchstack">', stack[i:]])
                 else:
-                    stack = stack[:i] + '<span class="matchstack">' + stack[i:k] \
-                        + '</span>' + stack[k:]
+                    stack = ''.join([stack[:i], '<span class="matchstack">', stack[i:k],
+                                     '</span>', stack[k:]])
             else:
                 stack = '<span class="matchstack">{}</span>'.format(stack)
         elif status == self.FAIL:
             classes[idx['text']] = 'failtext'
         else:  # ERROR
-            stack += '<br/>\n"%s"' % self.err_msg()
+            stack += '<br/>\n"%s"' % self.err_msg().replace('<', '&lt;').replace('>', '&gt;')
             classes[idx['text']] = 'errortext'
         tpl = self.Snapshot(str(self.line_col[0]), str(self.line_col[1]),
                             stack, status, excerpt)  # type: Tuple[str, str, str, str, str]
@@ -401,34 +420,24 @@ class HistoryRecord:
 
     @property
     def stack(self) -> str:
-        # return "->".join(tag_name for tag_name, _ in self.call_stack)
-        short_stack = []
-        anonymous_tail = True
-        for tag_name, _ in reversed(self.call_stack):
-            if tag_name[:1] == ':':
-                if anonymous_tail and tag_name != ":Forward":
-                    short_stack.append(tag_name)
-            else:
-                short_stack.append(tag_name)
-                anonymous_tail = False
-        return "->".join(reversed(short_stack))
+        return callstack_as_str(self.call_stack)
 
     @property
     def status(self) -> str:
         if self.errors:
             return self.ERROR + ": " + ', '.join(str(e.code) for e in self.errors)
-        elif self.node.tag_name in (NONE_TAG, ZOMBIE_TAG):
+        elif self.node.name in (NONE_TAG, ZOMBIE_TAG):
             return self.FAIL
-        elif self.node.tag_name == EMPTY_PTYPE:
+        elif self.node.name == EMPTY_PTYPE:
             return self.DROP
         else:
             return self.MATCH
-        # return self.FAIL if self.node is None or self.node.tag_name == ZOMBIE_TAG else \
+        # return self.FAIL if self.node is None or self.node.name == ZOMBIE_TAG else \
         #     ('"%s"' % self.err_msg()) if self.errors else self.MATCH
 
     @property
     def excerpt(self):
-        if self.node.tag_name not in (NONE_TAG, ZOMBIE_TAG) and not self.errors:
+        if self.node.name not in (NONE_TAG, ZOMBIE_TAG) and not self.errors:
             excerpt = abbreviate_middle(str(self.node), 40)
         else:
             s = self.text
@@ -443,7 +452,7 @@ class HistoryRecord:
 
     @property
     def remaining(self) -> int:
-        return len(self.text) - len(self.node)
+        return len(self.text) - self.node.strlen()
 
     @staticmethod
     def last_match(history: List['HistoryRecord']) -> Union['HistoryRecord', None]:
@@ -528,7 +537,7 @@ def log_ST(syntax_tree, log_file_name) -> bool:
     return False
 
 
-def log_parsing_history(grammar, log_file_name: str = '', html: bool = True) -> bool:
+def log_parsing_history(grammar, log_file_name: str = '', as_html: bool = True) -> bool:
     """
     Writes a log of the parsing history of the most recently parsed document, if
     logging is turned on. Returns True, if that was the case and writing the
@@ -540,19 +549,19 @@ def log_parsing_history(grammar, log_file_name: str = '', html: bool = True) ->
         log_file_name (str):  The (base-)name of the log file to be written.
             If no name is given (default), then the class name of the grammar
             object will be used.
-        html (bool):  If true (default), the log will be output as html-Table,
+        as_html (bool):  If true (default), the log will be output as html-Table,
             otherwise as plain test. (Browsers might take a few seconds or
             minutes to display the table for long histories.)
     """
     def write_log(history: List[str], log_name: str) -> None:
-        htm = '.html' if html else ''
+        htm = '.html' if as_html else ''
         path = os.path.join(log_dir(), log_name + "_parser.log" + htm)
         if os.path.exists(path):
             os.remove(path)
             # print('WARNING: Log-file "%s" already existed and was deleted.' % path)
         if history:
             with open(path, "w", encoding="utf-8") as f:
-                if html:
+                if as_html:
                     f.write(HistoryRecord.HTML_LEAD_IN + '\n')
                     f.writelines(history)
                     f.write('\n</table>\n' + HistoryRecord.HTML_LEAD_OUT)
@@ -564,7 +573,7 @@ def log_parsing_history(grammar, log_file_name: str = '', html: bool = True) ->
         table every 100 rows to allow browser to speed up rendering.
         Does this really work...?"""
         log.append(line)
-        if html and len(log) % 50 == 0:
+        if as_html and len(log) % 50 == 0:
             log.append('\n'.join(['</table>\n<table>', HistoryRecord.COLGROUP]))
 
     if not is_logging():
@@ -589,7 +598,7 @@ def log_parsing_history(grammar, log_file_name: str = '', html: bool = True) ->
     history.append(lead_in)
 
     for record in grammar.history__[-LOG_SIZE_THRESHOLD:]:
-        line = record.as_html_tr() if html else (str(record) + '\n')
+        line = record.as_html_tr() if as_html else (str(record) + '\n')
         append_line(history, line)
 
     write_log(history, log_file_name)
