@@ -54,11 +54,11 @@ from DHParser import start_logging, suspend_logging, resume_logging, is_filename
     ZeroOrMore, Forward, NegativeLookahead, Required, CombinedParser, mixin_comment, \
     compile_source, grammar_changed, last_value, matching_bracket, PreprocessorFunc, is_empty, \
     remove_if, Node, TransformerCallable, TransformationDict, transformation_factory, traverse, \
-    remove_children_if, move_adjacent, normalize_whitespace, is_anonymous, matches_re, \
+    remove_children_if, move_fringes, normalize_whitespace, is_anonymous, matches_re, \
     reduce_single_child, replace_by_single_child, replace_or_reduce, remove_whitespace, \
     replace_by_children, remove_empty, remove_tokens, flatten, all_of, any_of, \
     merge_adjacent, collapse, collapse_children_if, transform_content, WHITESPACE_PTYPE, \
-    TOKEN_PTYPE, remove_children, remove_content, remove_brackets, change_tag_name, \
+    TOKEN_PTYPE, remove_children, remove_content, remove_brackets, change_name, \
     remove_anonymous_tokens, keep_children, is_one_of, not_one_of, has_content, apply_if, peek, \
     remove_anonymous_empty, keep_nodes, traverse_locally, strip, lstrip, rstrip, \
     transform_content, replace_content_with, forbid, assert_content, remove_infix_operator, \
@@ -66,7 +66,7 @@ from DHParser import start_logging, suspend_logging, resume_logging, is_filename
     get_config_value, node_maker, access_thread_locals, access_presets, PreprocessorResult, \
     finalize_presets, ErrorCode, RX_NEVER_MATCH, set_tracer, resume_notices_on, \
     trace_history, has_descendant, neg, has_ancestor, optional_last_value, insert, \
-    positions_of, replace_tag_names, add_attributes, delimit_children, merge_connected, \
+    positions_of, replace_child_names, add_attributes, delimit_children, merge_connected, \
     has_attr, has_parent, ThreadLocalSingletonFactory, Error, canonical_error_strings, \
     has_errors, ERROR, FATAL, set_preset_value, get_preset_value, NEVER_MATCH_PATTERN, \
     gen_find_include_func, preprocess_includes, make_preprocessor, chain_preprocessors, \
@@ -199,7 +199,7 @@ def parse_ts2python(document, start_parser = "root_parser__", *, complete_match=
 ts2python_AST_transformation_table = {
     # AST Transformations for the ts2python-grammar
     # "<": flatten,
-    ":Text": change_tag_name('TEXT')
+    ":Text": change_name('TEXT')
     # "*": replace_by_single_child
 }
 
@@ -207,7 +207,7 @@ ts2python_AST_transformation_table = {
 def ts2pythonTransformer() -> TransformerCallable:
     """Creates a transformation function that does not share state with other
     threads or processes."""
-    return partial(traverse, processing_table=ts2python_AST_transformation_table.copy())
+    return partial(traverse, transformation_table=ts2python_AST_transformation_table.copy())
 
 
 get_transformer = ThreadLocalSingletonFactory(ts2pythonTransformer, ident=1)
@@ -317,7 +317,7 @@ class ts2pythonCompiler(Compiler):
         result = super().compile(node)
         if isinstance(result, str):
             return result
-        raise TypeError(f"Compilation of {node.tag_name} yielded a result of "
+        raise TypeError(f"Compilation of {node.name} yielded a result of "
                         f"type {str(type(result))} and not str as expected!")
 
     def is_toplevel(self) -> bool:
@@ -336,7 +336,7 @@ class ts2pythonCompiler(Compiler):
         return None
 
     def finalize(self, python_code: Any) -> Any:
-        if self.tree.tag_name == 'document':
+        if self.tree.name == 'document':
             code_blocks = [
                 f'# Generated by ts2python on {datetime.datetime.now()}\n',
                 GENERAL_IMPORTS, TYPEDDICT_IMPORTS, self.additional_imports]
@@ -358,7 +358,7 @@ class ts2pythonCompiler(Compiler):
 
     def on_document(self, node) -> str:
         return '\n\n'.join(self.compile(child) for child in node.children
-                           if child.tag_name != 'declaration')
+                           if child.name != 'declaration')
 
     def render_class_header(self, name: str, base_classes: str) -> str:
         optional_key_list = self.optional_keys.pop()
@@ -433,7 +433,7 @@ class ts2pythonCompiler(Compiler):
 
     def on_type_alias(self, node) -> str:
         alias = self.compile(node['identifier'])
-        if all(typ[0].tag_name in ('basic_type', 'literal') for typ in node.select('type')):
+        if all(typ[0].name in ('basic_type', 'literal') for typ in node.select('type')):
             self.basic_type_aliases.add(alias)
         self.obj_name.append(alias)
         if alias not in self.overloaded_type_names:
@@ -450,7 +450,7 @@ class ts2pythonCompiler(Compiler):
 
     def on_declarations_block(self, node) -> str:
         declarations = '\n'.join(self.compile(nd) for nd in node
-                                 if nd.tag_name == 'declaration')
+                                 if nd.name == 'declaration')
         return declarations or "pass"
 
     def on_declaration(self, node) -> str:
@@ -516,8 +516,8 @@ class ts2pythonCompiler(Compiler):
             else:
                 preface = ''
             if self.use_literal_type and \
-                    any(nd[0].tag_name == 'literal' for nd in node.children):
-                assert all(nd[0].tag_name == 'literal' for nd in node.children)
+                    any(nd[0].name == 'literal' for nd in node.children):
+                assert all(nd[0].name == 'literal' for nd in node.children)
                 return f"Literal[{', '.join(union)}]"
             elif self.use_type_union:
                 return preface + '|'.join(union)
@@ -529,7 +529,7 @@ class ts2pythonCompiler(Compiler):
     def on_type(self, node) -> str:
         assert len(node.children) == 1
         typ = node[0]
-        if typ.tag_name == 'declarations_block':
+        if typ.name == 'declarations_block':
             self.local_classes.append([])
             self.optional_keys.append([])
             decls = self.compile(typ)
@@ -537,8 +537,8 @@ class ts2pythonCompiler(Compiler):
                              self.render_local_classes().replace('\n', '\n    '),
                              decls.replace('\n', '\n    ')])   # maybe add one '\n'?
             # return 'Dict'
-        elif typ.tag_name == 'literal':
-            literal_typ = typ[0].tag_name
+        elif typ.name == 'literal':
+            literal_typ = typ[0].name
             if self.use_literal_type:
                 return self.compile(typ)
             elif literal_typ == 'array':
@@ -574,8 +574,8 @@ class ts2pythonCompiler(Compiler):
         if name in self.known_types:  return ''
         self.known_types.add(name)
         save = self.strip_type_from_const
-        if all(child.tag_name == 'const' for child in node.children[1:]):
-            if all(nd['literal'][0].tag_name == 'integer'
+        if all(child.name == 'const' for child in node.children[1:]):
+            if all(nd['literal'][0].name == 'integer'
                    for nd in node.select_children('const') if 'literal' in nd):
                 header = f'class {name}(IntEnum):'
             else:
@@ -594,7 +594,7 @@ class ts2pythonCompiler(Compiler):
 
     def on_enum(self, node) -> str:
         if self.use_enums:
-            if all(nd['literal'][0].tag_name == 'integer' for
+            if all(nd['literal'][0].name == 'integer' for
                    nd in node.select_children('item')):
                 base_class = '(IntEnum)'
             else:
@@ -680,7 +680,7 @@ class ts2pythonCompiler(Compiler):
     def on_array_of(self, node) -> str:
         assert len(node.children) == 1
         name = self.compile(node[0])
-        if node[0].tag_name == 'identifier' and name not in self.known_types:
+        if node[0].name == 'identifier' and name not in self.known_types:
             name = "'" + name + "'"
         return 'List[' + name + ']'
 
