@@ -43,11 +43,13 @@ from typing import Dict, List, Union, Deque, cast
 from DHParser.configuration import get_config_value
 from DHParser.compile import run_pipeline
 from DHParser.error import Error, is_error, PARSER_LOOKAHEAD_MATCH_ONLY, \
-    PARSER_LOOKAHEAD_FAILURE_ONLY, MANDATORY_CONTINUATION_AT_EOF, AUTORETRIEVED_SYMBOL_NOT_CLEARED
+    PARSER_LOOKAHEAD_FAILURE_ONLY, MANDATORY_CONTINUATION_AT_EOF, \
+    MANDATORY_CONTINUATION_AT_EOF_NON_ROOT, CAPTURE_STACK_NOT_EMPTY_NON_ROOT_ONLY, \
+    AUTOCAPTURED_SYMBOL_NOT_CLEARED_NON_ROOT
 from DHParser.log import is_logging, clear_logs, local_log_dir, log_parsing_history
 from DHParser.parse import Lookahead
 from DHParser.server import RX_CONTENT_LENGTH, RE_DATA_START, JSONRPC_HEADER_BYTES
-from DHParser.syntaxtree import Node, RootNode, deserialize, flatten_sxpr, ZOMBIE_TAG, EMPTY_PTYPE
+from DHParser.nodetree import Node, RootNode, deserialize, flatten_sxpr, ZOMBIE_TAG, EMPTY_PTYPE
 from DHParser.trace import set_tracer, all_descendants, trace_history
 from DHParser.transform import traverse, remove_children
 from DHParser.toolkit import load_if_file, re, re_find, concurrent_ident, instantiate_executor
@@ -90,7 +92,7 @@ RX_COMMENT = re.compile(r'\s*[#;].*(?:\n|$)')
 
 
 def normalize_code(testcode: str, full_normalization: bool=False) -> str:
-    """Removes leading and trailing empty lines (if full_normalization ist True)
+    """Removes leading and trailing empty lines (if full_normalization is True)
     and leading indentation (always) from multiline text. Single line text
     will be returned unchanged.
     """
@@ -104,6 +106,8 @@ def normalize_code(testcode: str, full_normalization: bool=False) -> str:
             if line:
                 indent = min(indent, len(line) - len(line.lstrip()))
         if indent > 0 and indent != sys.maxsize:
+            if lines[0].strip() and lines[0][0:1] not in ('', ' ') and indent > 4:
+                indent = min(4, max(indent - 4, 4))
             for i in range(1, len(lines)):
                 lines[i] = lines[i][indent:]
         if full_normalization:
@@ -322,8 +326,9 @@ def get_report(test_unit) -> str:
 POSSIBLE_ARTIFACTS = frozenset((
     PARSER_LOOKAHEAD_MATCH_ONLY,
     PARSER_LOOKAHEAD_FAILURE_ONLY,
-    MANDATORY_CONTINUATION_AT_EOF,
-    AUTORETRIEVED_SYMBOL_NOT_CLEARED
+    MANDATORY_CONTINUATION_AT_EOF_NON_ROOT,
+    CAPTURE_STACK_NOT_EMPTY_NON_ROOT_ONLY,
+    AUTOCAPTURED_SYMBOL_NOT_CLEARED_NON_ROOT
 ))
 
 TEST_ARTIFACT = "__TEST_ARTIFACT__"
@@ -355,10 +360,10 @@ def grammar_unit(test_unit, parser_factory, transformer_factory, report='REPORT'
     output = []
 
     def write(s):
-        nonlocal output
         """Append string `s` to output. The purpose is to defer printing to
         stdout in order to avoid muddled output when several unit tests run
         at the same time."""
+        nonlocal output
         output.append(s)
 
     def clean_key(k):
@@ -404,27 +409,24 @@ def grammar_unit(test_unit, parser_factory, transformer_factory, report='REPORT'
         """
         if not get_config_value('test_suppress_lookahead_failures'):
             return False
-        raw_errors = cast(RootNode, syntax_tree).errors_sorted
-        is_artifact = ({e.code for e in raw_errors}
-                       <= {PARSER_LOOKAHEAD_FAILURE_ONLY,
-                           AUTORETRIEVED_SYMBOL_NOT_CLEARED,
-                           PARSER_LOOKAHEAD_MATCH_ONLY}
+        raw_errors = cast(RootNode, syntax_tree).errors
+        is_artifact = ({e.code for e in raw_errors} <= POSSIBLE_ARTIFACTS
                        or (len(raw_errors) == 1
                            and (raw_errors[-1].code == PARSER_LOOKAHEAD_MATCH_ONLY
                                 #  case 2:  mandatory lookahead failure at end of text
                                 or raw_errors[-1].code == MANDATORY_CONTINUATION_AT_EOF)))
         if is_artifact:
             # don't remove zombie node with error message at the end
-            # but change it's tag_name to indicate that it is an artifact!
-            for parent in syntax_tree.select_if(lambda node: any(child.tag_name == ZOMBIE_TAG
+            # but change it's name to indicate that it is an artifact!
+            for parent in syntax_tree.select_if(lambda node: any(child.name == ZOMBIE_TAG
                                                                  for child in node.children),
                                                 include_root=True, reverse=True):
                 zombie = parent.pick_child(ZOMBIE_TAG)
-                zombie.tag_name = TEST_ARTIFACT
+                zombie.name = TEST_ARTIFACT
                 zombie.result = 'Artifact can be ignored. Be aware, though, that also the ' \
                                 'tree structure may not be the same as in a non-testing ' \
                                 'environment, when a testing artifact has occurred!'
-                # parent.result = tuple(c for c in parent.children if c.tag_name != ZOMBIE_TAG)
+                # parent.result = tuple(c for c in parent.children if c.name != ZOMBIE_TAG)
                 break
         return is_artifact
 
@@ -436,7 +438,7 @@ def grammar_unit(test_unit, parser_factory, transformer_factory, report='REPORT'
                 if errata:  errata[-1] = errata[-1].rstrip('\n')
                 errata.append('\t' + '\n\t'.join(
                     str(msg).replace('\n', '\n\t\t') for msg in test_errors))
-                errata.append('\n\n')
+                # errata.append('\n\n')  # leads to wrong error count!!!
 
     for parser_name, tests in test_unit.items():
         # if not get_config_value('test_parallelization'):
@@ -493,11 +495,11 @@ def grammar_unit(test_unit, parser_factory, transformer_factory, report='REPORT'
             tests.setdefault('__cst__', {})[test_name] = cst
             # errors = []  # type: List[Error]
             if is_error(cst.error_flag) and not lookahead_artifact(cst):
-                errors = [e for e in cst.errors_sorted if e.code not in POSSIBLE_ARTIFACTS]
+                errors = [e for e in cst.errors if e.code not in POSSIBLE_ARTIFACTS]
                 errata.append('Match test "%s" for parser "%s" failed:'
-                              '\nExpr.:  %s\n\n%s\n\n' %
+                              '\n\tExpr.:  %s\n\t%s' %
                               (test_name, parser_name, md_codeblock(test_code),
-                               '\n'.join(str(m).replace('\n', '\n') for m in errors)))
+                               '\n'.join(str(m) for m in errors)))
             if "ast" in tests or report or transformation_stages or show:
                 ast = copy.deepcopy(cst) if 'cst' in tests or str(test_name).find('*') >= 0 \
                       else cst
@@ -646,10 +648,11 @@ def grammar_unit(test_unit, parser_factory, transformer_factory, report='REPORT'
                 # write parsing-history log only in case of test-failure
                 if is_logging() and track_history:
                     with local_log_dir('./LOGS'):
-                        log_parsing_history(parser, "fail_%s_%s.log" % (parser_name, test_name))
+                        tname = test_name.replace('*', '')
+                        log_parsing_history(parser, f"fail_{parser_name}_{tname}.log")
             if cst.error_flag:
                 tests.setdefault('__msg__', {})[test_name] = \
-                    "\n".join(str(e) for e in cst.errors_sorted)
+                    "\n".join(str(e) for e in cst.errors)
             if verbose:
                 infostr = '    fail-test  "' + test_name + '" ... '
                 write(infostr + ("OK" if len(errata) == errflag else "FAIL"))
@@ -763,7 +766,7 @@ def grammar_suite(directory, parser_factory, transformer_factory,
     if error_report:
         # if verbose:
         #     print("\nFAILURE! %i error%s found!\n" % (err_N, 's' if err_N > 1 else ''))
-        return ('Test suite "%s" revealed %s error%s:\n\n'
+        return ('Test suite "%s" revealed %s error%s:\n'
                 % (directory, err_N, 's' if err_N > 1 else '') + '\n'.join(error_report))
     if verbose:
         print("\nSUCCESS! All tests passed :-)\n")
@@ -780,6 +783,8 @@ def grammar_suite(directory, parser_factory, transformer_factory,
 RX_DEFINITION_OR_SECTION = re.compile(r'(?:^|\n)[ \t]*(\w+(?=[ \t]*=)|#:.*(?=\n|$|#))')
 SymbolsDictType = Dict[str, List[str]]
 
+ALL_SYMBOLS = 'ALL_SYMBOLS'
+
 
 def extract_symbols(ebnf_text_or_file: str) -> SymbolsDictType:
     r"""
@@ -828,7 +833,7 @@ def extract_symbols(ebnf_text_or_file: str) -> SymbolsDictType:
             deflist = '#: ALL'
     symbols = collections.OrderedDict()  # type: SymbolsDictType
     if deflist[0][:2] != '#:':
-        curr_section = ''
+        curr_section = ALL_SYMBOLS
         symbols[curr_section] = []
     for df in deflist:
         if df[:2] == '#:':
@@ -868,7 +873,7 @@ def create_test_templates(symbols_or_ebnf: Union[str, SymbolsDictType],
         save = os.getcwd()
         os.chdir(path)
         keys = reversed(list(symbols.keys()))
-        existing_tests = {fname[3:]: fname for fname in os.listdir()}
+        existing_tests = {fname[3:]: fname for fname in os.listdir() if os.path.isfile(fname)}
         for i, k in enumerate(keys):
             filename = '{num:0>2}_test_{section}'.format(num=i + 1, section=k) + fmt
             if not os.path.exists(filename):
@@ -877,7 +882,7 @@ def create_test_templates(symbols_or_ebnf: Union[str, SymbolsDictType],
                     print(f'Renaming test file "{old_name}" to "{filename}"')
                     os.rename(old_name, filename)
                     existing_tests[filename[3:]] = filename
-                else:
+                elif k is not ALL_SYMBOLS or not existing_tests:
                     print('Creating test file template "{name}".'.format(name=filename))
                     with open(filename, 'w', encoding='utf-8') as f:
                         for sym in symbols[k]:
@@ -1027,7 +1032,8 @@ def runner(tests, namespace, profile=False):
 
 
 def run_file(fname):
-    if fname.lower().startswith('test_') and fname.endswith('.py'):
+    f_lower = fname.lower()
+    if f_lower.startswith('test_') and f_lower.endswith('.py'):
         print("RUNNING " + fname)
         # print('\nRUNNING UNIT TESTS IN: ' + fname)
         exec('import ' + fname[:-3])
@@ -1043,7 +1049,9 @@ def run_path(path):
         with instantiate_executor(get_config_value('test_parallelization') and len(files) > 1,
                                   concurrent.futures.ProcessPoolExecutor) as pool:
             for f in files:
-                results.append(pool.submit(run_file, f))
+                f_lower = f.lower()
+                if f_lower.startswith('test_') and f_lower.endswith('.py'):
+                    results.append(pool.submit(run_file, f))
                 # run_file(f)  # for testing!
             concurrent.futures.wait(results)
             for r in results:
