@@ -35,6 +35,7 @@ import multiprocessing
 import os
 import sys
 import threading
+import traceback
 
 assert sys.version_info >= (3, 5, 3), "DHParser requires at least Python-Version 3.5.3!"
 
@@ -50,7 +51,7 @@ except ImportError:
 
 import typing
 from typing import Any, Iterable, Sequence, Set, AbstractSet, Union, Dict, List, Tuple, \
-    Optional, Type
+    Optional, Type, Callable
 try:
     from typing import Protocol
 except ImportError:
@@ -91,6 +92,7 @@ __all__ = ('typing',
            'escape_re',
            'escape_ctrl_chars',
            'is_filename',
+           'is_html_name',
            'relative_path',
            'split_path',
            'concurrent_ident',
@@ -99,6 +101,7 @@ __all__ = ('typing',
            'escape_formatstr',
            'as_identifier',
            'as_list',
+           'as_tuple',
            'first',
            'last',
            'NOPE',
@@ -106,7 +109,7 @@ __all__ = ('typing',
            'linebreaks',
            'line_col',
            'text_pos',
-           'lstrip_docstring',
+           'normalize_docstring',
            'issubtype',
            'isgenerictype',
            'load_if_file',
@@ -125,7 +128,8 @@ __all__ = ('typing',
            'sane_parser_name',
            'normalize_circular_path',
            'DHPARSER_DIR',
-           'DHPARSER_PARENTDIR')
+           'deprecated',
+           'deprecation_warning')
 
 
 #######################################################################
@@ -149,7 +153,7 @@ def identity(x):
 
 
 if __name__ == "__main__":
-  global_id_counter = multiprocessing.Value('L', 0)
+    global_id_counter = multiprocessing.Value('L', 0)
 
 
 def gen_id() -> int:
@@ -192,6 +196,11 @@ def is_filename(strg: str) -> bool:
     #   and strg.select_if('*') < 0 and strg.select_if('?') < 0
 
 
+def is_html_name(url: str) -> bool:
+    """Returns True, if url ends with .htm or .html"""
+    return url[-5:].lower() == '.html' or url[-4:].lower() == '.htm'
+
+
 @cython.locals(i=cython.int, L=cython.int)
 def relative_path(from_path: str, to_path: str) -> str:
     """Returns the relative path in order to open a file from
@@ -213,14 +222,14 @@ def relative_path(from_path: str, to_path: str) -> str:
     return os.path.normpath(from_path[i:].count('/') * '../' + to_path[i:])
 
 
-def split_path(path: str) -> Tuple[str]:
+def split_path(path: str) -> Tuple[str, ...]:
     """Splits a filesystem path into its components. Other than
     os.path.split() it does not only split of the last part::
 
-    >>> split_path('a/b/c')
-    ('a', 'b', 'c')
-    >>> os.path.split('a/b/c')  # for comparison.
-    ('a/b', 'c')
+        >>> split_path('a/b/c')
+        ('a', 'b', 'c')
+        >>> os.path.split('a/b/c')  # for comparison.
+        ('a/b', 'c')
     """
     split = os.path.split(path)
     while split[0]:
@@ -275,6 +284,14 @@ def as_list(item_or_sequence) -> List[Any]:
     return [item_or_sequence]
 
 
+def as_tuple(item_or_sequence) -> List[Any]:
+    """Turns an arbitrary sequence or a single item into a tuple. In case of
+    a single item, the tuple contains this element as its sole item."""
+    if isinstance(item_or_sequence, Iterable):
+        return tuple(item_or_sequence)
+    return (item_or_sequence,)
+
+
 def first(item_or_sequence: Union[Sequence, Any]) -> Any:
     """Returns an item or the first item of a sequence of items."""
     if isinstance(item_or_sequence, Sequence):
@@ -291,6 +308,44 @@ def last(item_or_sequence: Union[Sequence, Any]) -> Any:
         return item_or_sequence
 
 
+DEPRECATION_WARNINGS_ISSUED: Set[str] = set()
+
+
+def deprecation_warning(message: str):
+    """Issues a deprecation warning. Makes sure that each message is only
+    called once."""
+    if message not in DEPRECATION_WARNINGS_ISSUED:
+        DEPRECATION_WARNINGS_ISSUED.add(message)
+        try:
+            raise DeprecationWarning(message)
+        except DeprecationWarning as e:
+            if get_config_value('deprecation_policy') == 'warn':
+                stacktrace = traceback.format_exc()
+                print(stacktrace)
+            else:
+                raise e
+
+
+def deprecated(message: str) -> Callable:
+    """Marks a function as deprecated and emits a deprecation
+    message on its first use::
+
+        >>> @deprecated('This function is deprecated!')
+        ... def bad():
+        ...     pass
+        >>> bad()
+        This function is deprecated!
+    """
+    assert isinstance(message, str)
+    def decorator(f: Callable) -> Callable:
+        @functools.wraps(f)
+        def wrapper(*args, **kwargs):
+            deprecation_warning(message)
+            return f(*args, **kwargs)
+        return wrapper
+    return decorator
+
+
 #######################################################################
 #
 # miscellaneous (DHParser-specific)
@@ -299,7 +354,7 @@ def last(item_or_sequence: Union[Sequence, Any]) -> Any:
 
 
 DHPARSER_DIR = os.path.dirname(os.path.abspath(__file__))
-DHPARSER_PARENTDIR = os.path.dirname(DHPARSER_DIR.rstrip('/'))
+# DHPARSER_PARENTDIR = os.path.dirname(DHPARSER_DIR.rstrip('/').rstrip('\\'))
 
 
 def sane_parser_name(name) -> bool:
@@ -311,8 +366,8 @@ def sane_parser_name(name) -> bool:
     return name and name[:2] != '__' and name[-2:] != '__'
 
 
-def normalize_circular_path(path: Union[Tuple[str], AbstractSet[Tuple[str]]]) \
-        -> Union[Tuple[str], Set[Tuple[str]]]:
+def normalize_circular_path(path: Union[Tuple[str, ...], AbstractSet[Tuple[str, ...]]]) \
+        -> Union[Tuple[str, ...], Set[Tuple[str, ...]]]:
     """Returns a normalized version of a `circular path` represented as
     a tuple or - if called with a set of paths instead of a single path
     - a set of normalized paths.
@@ -362,14 +417,14 @@ def re_find(s, r, pos=0, endpos=9223372036854775807):
     """
     if isinstance(r, (str, bytes)):
         if (pos, endpos) != (0, 9223372036854775807):
-            rx = re.compile(r)
+            r = re.compile(r)
         else:
             try:
                 m = next(re.finditer(r, s))
                 return m
             except StopIteration:
                 return None
-    elif r:
+    if r:
         try:
             m = next(r.finditer(s, pos, endpos))
             return m
@@ -405,20 +460,30 @@ def escape_ctrl_chars(strg: str) -> str:
     return s
 
 
-def lstrip_docstring(docstring: str) -> str:
+def normalize_docstring(docstring: str) -> str:
     """
-    Strips leading whitespace from a docstring.
+    Strips leading indentation as well as leading
+    and trailing empty lines from a docstring.
     """
-
     lines = docstring.replace('\t', '    ').split('\n')
+    # determine indentation
     indent = 255  # highest integer value
     for line in lines[1:]:
         stripped = line.lstrip()
         if stripped:  # ignore empty lines
             indent = min(indent, len(line) - len(stripped))
-    if indent >= 255:
-        indent = 0
-    return '\n'.join([lines[0]] + [line[indent:] for line in lines[1:]])
+    if indent >= 255:  indent = 0
+    # remove trailing empty lines
+    while lines and not lines[-1].strip(): lines.pop()
+    if lines:
+        if lines[0].strip():
+            lines = [lines[0]] + [line[indent:] for line in lines[1:]]
+        else:
+            lines = [line[indent:] for line in lines[1:]]
+        # remove any empty lines at the beginning
+        while lines and not lines[0].strip(): del lines[0]
+        return '\n'.join(lines)
+    return ''
 
 
 @cython.locals(max_length=cython.int, length=cython.int, a=cython.int, b=cython.int)
