@@ -28,54 +28,72 @@ from typing import Tuple, Optional, List, Iterable, Union
 
 from DHParser.error import Error, RESUME_NOTICE, RECURSION_DEPTH_LIMIT_HIT
 from DHParser.stringview import StringView
-from DHParser.syntaxtree import Node, REGEXP_PTYPE, TOKEN_PTYPE, WHITESPACE_PTYPE
-from DHParser.log import HistoryRecord
-from DHParser.parse import Grammar, Parser, ParserError, ParseFunc
+from DHParser.nodetree import Node, REGEXP_PTYPE, TOKEN_PTYPE, WHITESPACE_PTYPE
+from DHParser.log import HistoryRecord, callstack_as_str
+from DHParser.parse import Grammar, Parser, ParserError, ParseFunc, ContextSensitive
 from DHParser.toolkit import cython, line_col
 
 __all__ = ('trace_history', 'all_descendants', 'set_tracer',
-           'resume_notices_on')
+           'resume_notices_on', 'resume_notices_off')
+
+
+def symbol_name(parser: Parser, grammar: Grammar) -> str:
+    name = str(parser) if isinstance(parser, ContextSensitive) else parser.node_name
+    # name = parser.name
+    if name[:1] == ':':
+        name = grammar.associated_symbol__(parser).pname + '->' + name
+    return name
 
 
 @cython.locals(location=cython.int, loc=cython.int, delta=cython.int, cs_len=cython.int,
                i=cython.int, L=cython.int)
 def trace_history(self: Parser, text: StringView) -> Tuple[Optional[Node], StringView]:
     grammar = self._grammar  # type: Grammar
+    if not grammar.history_tracking__:
+        try:
+            node, rest = self._parse(text)  # <===== call to the actual parser!
+        except ParserError as pe:
+            raise pe
+        return node, rest
+
     location = grammar.document_length__ - text._len  # type: int
 
-    if grammar.most_recent_error__:
+    mre: Optional[ParserError] = grammar.most_recent_error__
+    if mre is not None and location >= mre.error.pos:
         # add resume notice (mind that skip notices are added by
         # `parse.MandatoryElementsParser.mandatory_violation()`
-        mre = grammar.most_recent_error__  # type: ParserError
         if mre.error.code == RECURSION_DEPTH_LIMIT_HIT:
             return mre.node, text
 
         grammar.most_recent_error__ = None
         errors = [mre.error]  # type: List[Error]
         text_ = grammar.document__[mre.error.pos:]
-        lc = line_col(grammar.document_lbreaks__, mre.error.pos)
-        resume_pos = self.grammar.document_length__ - len(text)
-        target = text if len(text) <= 10 else text[:7] + '...'
-
-        resumers = [grammar.call_stack__[-1][0]]
-        i = 2;  L = len(grammar.call_stack__)
-        while resumers[-1][:1] == ':' and i <= L:
-            resumers.append(grammar.call_stack__[-i][0])
-            i += 1
-        resumer = '->'.join(reversed(resumers))
+        orig_lc = line_col(grammar.document_lbreaks__, mre.error.pos)
+        orig_snippet= text_ if len(text_) <= 10 else text_[:7] + '...'
+        # orig_snippet = orig_rest if len(orig_rest) <= 10 else orig_rest[:7] + '...'
+        target_pos = grammar.document_length__ - len(text)
+        target_lc = line_col(grammar.document_lbreaks__, target_pos)
+        target_text_ = grammar.document__[target_pos:]
+        target_snippet = target_text_ if len(target_text_) <= 10 else target_text_[:7] + '...'
+        # target = text if len(text) <= 10 else text[:7] + '...'
 
         if mre.first_throw:
-            origin = mre.node.tag_name
-            if origin[:1] == ':':
-                origin = grammar.associated_symbol__(mre.parser).tag_name + '->' + origin
+            # origin = callstack_as_str(mre.callstack_snapshot, depth=3)
+            # resumer = callstack_as_str(grammar.call_stack__, depth=3)
+            origin = symbol_name(mre.parser, grammar)
+            resumer = symbol_name(self, grammar)
             notice = Error(  # resume notice
-                'Resuming from parser "{}" at position {}:{} with parser "{}": {}'
-                .format(origin, *lc, resumer, repr(target)),
-                resume_pos, RESUME_NOTICE)
+                'Resuming from {} at {}:{} {} with {} at {}:{} {}'
+                .format(origin, *orig_lc, repr(orig_snippet),
+                        resumer, *target_lc, repr(target_snippet)),
+                location, RESUME_NOTICE)
         else:
+            origin = symbol_name(self, grammar)
             notice = Error(  # skip notice
-                'Skipping from position {}:{} within parser {}: {}'
-                .format(*lc, resumer, repr(target)), resume_pos, RESUME_NOTICE)
+                'Skipping from {}:{} {} within {} to {}:{} {}'
+                .format(*orig_lc, repr(orig_snippet), origin,
+                        *target_lc, repr(target_snippet)),
+                location, RESUME_NOTICE)
         if grammar.resume_notices__:
             grammar.tree__.add_error(mre.node, notice)
         errors.append(notice)
@@ -84,8 +102,8 @@ def trace_history(self: Parser, text: StringView) -> Tuple[Optional[Node], Strin
             line_col(grammar.document_lbreaks__, mre.error.pos), errors))
 
     grammar.call_stack__.append(
-        (((' ' + self.repr) if self.tag_name in (REGEXP_PTYPE, TOKEN_PTYPE, ":Retrieve", ":Pop")
-          else (self.pname or self.tag_name)), location))  # ' ' added to avoid ':' as first char!
+        (((' ' + self.repr) if self.node_name in (REGEXP_PTYPE, TOKEN_PTYPE, ":Retrieve", ":Pop")
+          else (self.pname or self.node_name)), location))  # ' ' added to avoid ':' as first char!
     grammar.moving_forward__ = True
 
     try:
@@ -100,13 +118,13 @@ def trace_history(self: Parser, text: StringView) -> Tuple[Optional[Node], Strin
             # TODO: get the call stack from when the error occurred, here
             nd = fe.node
             grammar.history__.append(
-                HistoryRecord(grammar.call_stack__, nd, fe.rest[len(nd):], lc, [fe.error]))
+                HistoryRecord(grammar.call_stack__, nd, fe.rest[nd.strlen():], lc, [fe.error]))
         grammar.call_stack__.pop()
         raise pe
 
     # Mind that memoized parser calls will not appear in the history record!
     # Don't track returning parsers except in case an error has occurred!
-    if ((self.tag_name != WHITESPACE_PTYPE)
+    if ((self.node_name != WHITESPACE_PTYPE)
         and (grammar.moving_forward__
              or (not self.disposable
                  and (node
@@ -114,7 +132,7 @@ def trace_history(self: Parser, text: StringView) -> Tuple[Optional[Node], Strin
         # record history
         # TODO: Make dropping insignificant whitespace from history configurable
         delta = text._len - rest._len
-        hnd = Node(node.tag_name, text[:delta]).with_pos(location) if node else None
+        hnd = Node(node.name, text[:delta]).with_pos(location) if node else None
         lc = line_col(grammar.document_lbreaks__, location)
         record = HistoryRecord(grammar.call_stack__, hnd, rest, lc, [])
         cs_len = len(record.call_stack)
@@ -170,3 +188,8 @@ def resume_notices_on(grammar: Grammar):
     grammar.history_tracking__ = True
     grammar.resume_notices__ = True
     set_tracer(grammar, trace_history)
+
+
+def resume_notices_off(grammar: Grammar):
+    """Turns off resume-notices as well as history tracking!"""
+    set_tracer(grammar, None)
