@@ -26,7 +26,7 @@ import sys
 scriptpath = os.path.dirname(__file__) or '.'
 sys.path.append(os.path.abspath(os.path.join(scriptpath, '..')))
 
-from DHParser.syntaxtree import parse_sxpr
+from DHParser.nodetree import parse_sxpr, Node
 from DHParser.compile import Compiler
 
 
@@ -38,9 +38,9 @@ class SerializingTestCompiler(Compiler):
     def serialize(self, node):
         if node.children:
             content = [self.compile(child) for child in node.children]
-            return ' '.join(['(' + node.tag_name] + content) + ')'
+            return ' '.join(['(' + node.name] + content) + ')'
         else:
-            return '(' + node.tag_name + ' ' + node.content + ')'
+            return '(' + node.name + ' ' + node.content + ')'
 
     def on_A(self, node):
         return self.serialize(node)
@@ -61,6 +61,15 @@ class SerializingTestCompiler(Compiler):
         return self.serialize(node)
 
 
+class AttrTestCompiler(Compiler):
+    def attr_tic(self, node, value):
+        if node.children:
+            node.result = node.children + (Node('tic', f' tic: {value}'),)
+        else:
+            node.result = node.content + f' tic: {value}'
+        del node.attr['tic']
+
+
 class TestCompilerClass:
     original = parse_sxpr('(A (B "1") (C (D (E "2") (F "3"))))')
 
@@ -102,6 +111,13 @@ class TestCompilerClass:
                 "Incorrect Error Message: " + str(e)
             pass
 
+    def test_attribute_visitory(self):
+        attr_tree = parse_sxpr('(A (B `(tic "ticced :-)") "1") '
+                               '(C `(tic "ticced :-(") (D (E "2") (F "3"))))')
+        compiler = AttrTestCompiler()
+        s = compiler.compile(attr_tree)
+        assert str(s) == '1 tic: ticced :-)23 tic: ticced :-('
+
 
 if __name__ == "__main__":
     from DHParser.testing import runner
