@@ -28,7 +28,7 @@ scriptpath = os.path.dirname(__file__) or '.'
 sys.path.append(os.path.abspath(os.path.join(scriptpath, '..')))
 scriptpath = os.path.abspath(scriptpath)
 
-from DHParser.toolkit import compile_python_object, re, DHPARSER_PARENTDIR, \
+from DHParser.toolkit import compile_python_object, re, \
     normalize_circular_path
 from DHParser.preprocess import nil_preprocessor
 from DHParser import compile_source, INFINITE, Interleave
@@ -36,12 +36,15 @@ from DHParser.configuration import get_config_value, set_config_value
 from DHParser.error import has_errors, MANDATORY_CONTINUATION, PARSER_STOPPED_BEFORE_END, \
     REDEFINED_DIRECTIVE, UNUSED_ERROR_HANDLING_WARNING, AMBIGUOUS_ERROR_HANDLING, \
     REORDERING_OF_ALTERNATIVES_REQUIRED, BAD_ORDER_OF_ALTERNATIVES, UNCONNECTED_SYMBOL_WARNING, \
-    ERROR, WARNING
-from DHParser.syntaxtree import WHITESPACE_PTYPE
+    PEG_EXPRESSION_IN_DIRECTIVE_WO_BRACKETS, ERROR, WARNING, \
+    ZERO_LENGTH_CAPTURE_POSSIBLE_WARNING, canonical_error_strings
+from DHParser.nodetree import WHITESPACE_PTYPE, flatten_sxpr
 from DHParser.ebnf import get_ebnf_grammar, get_ebnf_transformer, EBNFTransform, \
-    EBNFDirectives, get_ebnf_compiler, compile_ebnf, DHPARSER_IMPORTS, parse_ebnf, transform_ebnf
+    EBNFDirectives, get_ebnf_compiler, compile_ebnf, DHPARSER_IMPORTS, parse_ebnf, \
+    transform_ebnf, HeuristicEBNFGrammar, ConfigurableEBNFGrammar
 from DHParser.dsl import CompilationError, compileDSL, create_parser, grammar_provider, raw_compileEBNF
-from DHParser.testing import grammar_unit, clean_report
+from DHParser.testing import grammar_unit, clean_report, unique_name
+from DHParser.trace import set_tracer, trace_history
 
 
 class TestDirectives:
@@ -142,6 +145,25 @@ class TestDirectives:
         st = parser('* Hund*')
         assert str(st) == "*Hund"
 
+    def test_drop_error_messages(self):
+        lang = r'''
+        @disposable = A
+        @drop = A, B
+        A = { B }
+        B = 'x'
+        '''
+        _, errors, _ = compile_ebnf(lang)
+        assert len(errors) == 1
+        assert errors[0].message.startswith('''Illegal value "B" for Directive "@ drop"!''')
+        assert errors[0].message.endswith('''where the "@disposable"-directive must precede the @drop-directive.''')
+        lang = r'''
+        @disposable = /[A]/
+        @drop = A, B
+        A = { B }
+        B = 'x'
+        '''
+        _, errors, _ = compile_ebnf(lang)
+        assert errors[0].message.find('or a string matching') >= 0
 
 
 class TestReservedSymbols:
@@ -226,6 +248,19 @@ class TestEBNFParser:
     def test_list(self):
         grammar_unit(self.cases, get_ebnf_grammar, get_ebnf_transformer, 'REPORT_TestEBNFParser')
 
+    def test_regex_start_end_marker_usable(self):
+        """Regular expression begin and end-marker really match only at
+        the document start and ending!"""
+        lang = '''
+        doc = BEGIN !END A !BEGIN !END B !BEGIN END
+        A = "A"
+        B = "B"
+        BEGIN = /^/
+        END = /$/
+        '''
+        parser = create_parser(lang)
+        result = parser('AB')
+        assert flatten_sxpr(result.as_sxpr()) == '(doc (BEGIN) (A "A") (B "B") (END))'
 
 
 class TestParserNameOverwriteBug:
@@ -450,8 +485,9 @@ class TestBoundaryCases:
             assert not has_errors(messages), "Unconnected rules should result in a warning, " \
                 "not an error: " + str(messages)
             grammar_src = result
-            grammar = compile_python_object(DHPARSER_IMPORTS + grammar_src,
-                                            r'get_(?:\w+_)?grammar$')()
+            grammar = compile_python_object(
+                DHPARSER_IMPORTS + grammar_src, #.format(dhparser_parentdir=repr('.')) + grammar_src,
+                r'get_(?:\w+_)?grammar$')()
         else:
             assert False, "EBNF compiler should warn about unconnected rules."
 
@@ -817,6 +853,35 @@ class TestErrorCustomizationErrors:
         assert "several strings" in str(result.errors), str(result.errors)
 
 
+class TestVariableCapture:
+    tree_grammar = '''@whitespace = horizontal
+        @disposable = EOF, LF, SAME_INDENT
+        @drop       = strings, whitespace, EOF, LF, SAME_INDENT
+        tree     = INDENT node DEDENT /\\\\s*/ EOF
+        node     = name [content]
+        content  = string | children
+        children = &(LF HAS_DEEPER_INDENT)
+                   LF INDENT § node { LF SAME_INDENT § node }
+                   !(LF HAS_DEEPER_INDENT) DEDENT
+        name = /\\\\w+/~
+        string   = '"' § /(?:\\\\\\\\"|[^"\\\\n])*/ '"' ~
+        INDENT            = / */
+        SAME_INDENT       = :INDENT § !/ /
+        HAS_DEEPER_INDENT = :INDENT / +/
+        DEDENT            = &:?INDENT
+        LF       = /\\\\n/
+        EOF      = !/./
+        '''
+
+    def test_zero_length_capture_warning(self):
+        pysrc, errors, _ = compile_ebnf(self.tree_grammar)
+        assert errors and errors[0].code == ZERO_LENGTH_CAPTURE_POSSIBLE_WARNING
+        alt_tree_grammar = self.tree_grammar.replace('INDENT            = / */',
+                                                     'INDENT            = / +/')
+        pysrc, errors, _ = compile_ebnf(alt_tree_grammar)
+        assert not errors
+
+
 class TestCustomizedResumeParsing:
     lang = r"""@ literalws = right
         @ alpha_resume = "BETA", "GAMMA"
@@ -825,10 +890,10 @@ class TestCustomizedResumeParsing:
         document = alpha [beta] gamma "."
         alpha = "ALPHA" abc
             abc = §"a" "b" "c"
-          beta = "BETA" (bac | bca)
+        beta = "BETA" (bac | bca)
             bac = "b" "a" §"c"
             bca = "b" "c" §"a"
-          gamma = "GAMMA" §(cab | cba)
+        gamma = "GAMMA" §(cab | cba)
             cab = "c" "a" §"b"
             cba = "c" "b" §"a"
         GAMMA_RE = /(?=GA\w+)/
@@ -841,7 +906,7 @@ class TestCustomizedResumeParsing:
         assert cst.error_flag
         assert cst.content == content
         assert cst.pick('alpha').content.startswith('ALPHA')
-        # because of resuming, there should be only on error message
+        # because of resuming, there should be only one error message
         assert len(cst.errors_sorted) == 1
 
         content = 'ALPHA acb BETA bad GAMMA cab .'
@@ -849,7 +914,7 @@ class TestCustomizedResumeParsing:
         assert cst.error_flag
         assert cst.content == content
         assert cst.pick('alpha').content.startswith('ALPHA')
-        # because of resuming, there should be only on error message
+        # because of resuming, there should be only one error message
         assert len(cst.errors_sorted) == 2
 
         content = 'ALPHA acb GAMMA cab .'
@@ -875,17 +940,18 @@ class TestCustomizedResumeParsing:
         grammar = grammar_provider(grammar_specification)()
         doc0 = """word no*word word"""
         st = grammar(doc0)
-        assert st.children and st.children[-1].tag_name == 'word'
+        assert st.children and st.children[-1].name == 'word'
         doc1 = """word no*word /* comment */ word"""
         st = grammar(doc1)
-        assert st.children and st.children[-1].tag_name == 'word'
+        assert st.children and st.children[-1].name == 'word'
         doc2 = """word no*word/* comment */word"""
         st = grammar(doc2)
-        assert st.children and st.children[-1].tag_name == 'word'
+        assert st.children and st.children[-1].name == 'word'
         doc3 = """word no*word/* comment1 */
                   /* comment2 */word"""
         st = grammar(doc3)
-        assert st.children and st.children[-1].tag_name == 'word'
+        assert st.children and st.children[-1].name == 'word'
+
 
 class TestCustomizedResumeParsing_with_Parsers:
     lang = r"""@ literalws = right
@@ -895,10 +961,10 @@ class TestCustomizedResumeParsing_with_Parsers:
         document = alpha [beta] gamma "."
         alpha = "ALPHA" abc
             abc = §"a" "b" "c"
-          beta = "BETA" (bac | bca)
+        beta = "BETA" (bac | bca)
             bac = "b" "a" §"c"
             bca = "b" "c" §"a"
-          gamma = "GAMMA" §(cab | cba)
+        gamma = "GAMMA" §(cab | cba)
             cab = "c" "a" §"b"
             cba = "c" "b" §"a"
         GAMMA_RE = /(?=GA\w+)/
@@ -912,7 +978,7 @@ class TestCustomizedResumeParsing_with_Parsers:
         assert cst.error_flag
         assert cst.content == content
         assert cst.pick('alpha').content.startswith('ALPHA')
-        # because of resuming, there should be only on error message
+        # because of resuming, there should be only one error message
         assert len(cst.errors_sorted) == 1
 
         content = 'ALPHA acb BETA bad GAMMA cab .'
@@ -920,7 +986,7 @@ class TestCustomizedResumeParsing_with_Parsers:
         assert cst.error_flag
         assert cst.content == content
         assert cst.pick('alpha').content.startswith('ALPHA')
-        # because of resuming, there should be only on error message
+        # because of resuming, there should be only one error message
         assert len(cst.errors_sorted) == 2
 
         content = 'ALPHA acb GAMMA cab .'
@@ -931,6 +997,100 @@ class TestCustomizedResumeParsing_with_Parsers:
         # because of resuming, there should be only one error message
         assert len(cst.errors_sorted) == 1, str(cst.errors_sorted)
 
+    def test_resume_with_expression(self):
+        lang = r"""@ literalws = right
+            @ alpha_resume =  ({ !`BETA` !`GAMMA` /./ })
+            @ beta_resume = GAMMA_RE
+            @ bac_resume = /(?=GA\w+)/
+            document = alpha [beta] gamma "."
+            alpha = "ALPHA" abc
+                abc = §"a" "b" "c"
+              beta = "BETA" (bac | bca)
+                bac = "b" "a" §"c"
+                bca = "b" "c" §"a"
+              gamma = "GAMMA" §(cab | cba)
+                cab = "c" "a" §"b"
+                cba = "c" "b" §"a"
+            GAMMA_RE = /(?=GA\w+)/
+            # ALPHA_RESUME = { !`BETA` !`GAMMA` /./ }
+            """
+        gr = grammar_provider(lang)()
+        # print(gr.python_src__)
+        content = 'ALPHA abc BETA bad GAMMA cab .'
+        cst = gr(content)
+        assert cst.error_flag
+        assert cst.content == content
+        assert cst.pick('alpha').content.startswith('ALPHA')
+        # because of resuming, there should be only on error message
+        assert len(cst.errors_sorted) == 1
+
+class TestCustomizedResumeParsing_with_autogenerated_Parsers:
+    lang = """@ literalws   = right
+    @ disposable  = /_\w+/
+    @ drop        = _EOF, whitespace, strings
+    _document = ~ [ list ] §_EOF
+    @list_skip = ({ list | /[^\[\],]*/ })
+    list     = "[" [_items] § "]"
+    @_items_skip = /(?=,)/, /(?=])/, /$/
+    _items   = _item { "," §_item }
+    _item    = number | list    
+    number   = `0` | /[1-9][0-9]*/
+    _EOF     =  !/./
+    """
+
+    def test_skip_expression_specification_error(self):
+        lang='''
+        @doc_skip = `` 'A' | 'B' | 'C'
+        doc = 'A' §'B' 'C'
+        '''
+        pysrc, errors, ast = compile_ebnf(lang, preserve_AST=True)
+        i = pysrc.find('root__ =') + 8
+        k = pysrc.find('\n', i)
+        assert pysrc[i:k].find('_skip_') < 0
+        assert len(errors) == 1
+        assert len([e for e in errors if e.code == PEG_EXPRESSION_IN_DIRECTIVE_WO_BRACKETS]) > 0
+
+    def test_skip_with_inline_parsers(self):
+        parser = create_parser(self.lang)
+        result = parser('[1, 2, A, [5, 6; [7, 8]], 8, ]')
+        assert all(err.code != PARSER_STOPPED_BEFORE_END for err in result.errors)
+        assert len(result.errors) == 3
+
+    def test_interaction_resume_capture(self):
+        lang='''
+        @ whitespace  = /\s*/
+        @ disposable  = EOF
+        @ drop        = EOF, whitespace, strings
+        
+        document = ~ element ~ §EOF
+        element  = STag content ETag
+        STag     = '<' TagName §'>'
+        @ETag_skip = ((`>` | :?TagName) /[^>]*/)
+        ETag     = '</' §::TagName '>'
+        TagName  = /\w+/
+        content  = [CharData] { (element | COMMENT__) [CharData] }
+        
+        CharData = /(?:(?!\]\]>)[^<&])+/
+        EOF      =  !/./
+        '''
+        pysrc, errors, ast = compile_ebnf(lang, preserve_AST=True)
+        assert not errors
+        i = pysrc.find('root__ =') + 8
+        k = pysrc.find('\n', i)
+        assert pysrc[i:k].find('_skip_') < 0
+
+        parser = create_parser(lang)
+        testdata = """
+        <doc>
+            <title>Heading</litle>
+            A few lines of Text
+        </doc>
+        """
+        set_tracer(parser, trace_history)
+        result = parser(testdata)
+        assert len(result.errors) == 1  # don't repeat error messages because of recursion!
+        assert len(parser.history__) == 33  # don't repeat history because of recursion!
+
 
 class TestInSeriesResume:
     def setup(self):
@@ -1089,11 +1249,70 @@ class TestAlternativeEBNFSyntax:
         code, errors, ast = compile_ebnf(ArithmeticEBNF, preserve_AST=True)
         assert not ast.error_flag, str(ast.errors)
         arithmetic_grammer = compile_python_object(
-            DHPARSER_IMPORTS.format(dhparser_parentdir=DHPARSER_PARENTDIR) + code)
+            DHPARSER_IMPORTS + code)  # .format(dhparser_parentdir=repr(DHPARSER_PARENTDIR)) + code)
         arithmetic_parser = arithmetic_grammer()
         st = arithmetic_parser('2 + 3 * (-4 + 1)')
         assert str(st) == "2+3*(-4+1)"
 
+    # def test_regex_heuristics(self):
+    #     gr = get_ebnf_grammar()
+    #     assert isinstance(gr, HeuristicEBNFGrammar)
+    #     result = gr(r' */', 'regex_heuristics')
+    #     print(result.as_sxpr())
+
+
+class TestConfigurableEBNF:
+    def setup(self):
+        self.save_variant = get_config_value('syntax_variant')
+        self.save_delim = get_config_value('delimiter_set')
+
+    def teardown(self):
+        set_config_value('syntax_variant', 'configurable')
+        set_config_value('delimiter_set', self.save_delim)
+        set_config_value('syntax_variant', self.save_variant)
+
+    def test_alt_syntax_config(self):
+        set_config_value('syntax_variant', 'configurable')
+        set_config_value('delimiter_set', {
+            'DEF': '<-',
+            'OR': '/',
+            'AND': '',
+            'ENDL': '',
+            'RNG_OPEN': '{',
+            'RNG_CLOSE': '}',
+            'RNG_DELIM': ',',
+            'TIMES': '*',
+            'RE_LEADIN': '<',
+            'RE_LEADOUT': '>',
+            'RE_CORE': r'(?:(?<!\\)\\(?:>)|[^>])*',
+            'CH_LEADIN': '0x'
+        })
+        parser = create_parser(
+            r'''@literalws=none
+            @disposable = Spacing, EOL, EOF
+            @drop = Spacing, EOL, EOF
+            Start   <- Spacing Expr EOL? EOF
+            Expr    <- Term ((PLUS / MINUS) Term)*
+            Term    <- Factor ((TIMES / DIVIDE) Factor)*
+            Factor  <- Sign* (LPAR Expr RPAR
+                             / INTEGER )
+            Sign    <- NEG / POS
+            INTEGER <- ( '0' / <[1-9][0-9]*> ) Spacing
+            PLUS    <- '+' Spacing
+            MINUS   <- '-' Spacing
+            TIMES   <- '*' Spacing
+            DIVIDE  <- '/' Spacing
+            LPAR    <- '(' Spacing
+            RPAR    <- ')' Spacing
+            NEG     <- '-' Spacing
+            POS     <- '+' Spacing
+            Spacing <- <[ \t\n\f\v\r]*>
+            EOL     <- '\r\n' / <[\n\r]>
+            EOF     <- !<.>
+            ''')
+        result = parser('3 + 4 * 7')
+        assert not result.errors
+
 
 class TestSyntaxExtensions:
     def setup(self):
@@ -1109,7 +1328,7 @@ class TestSyntaxExtensions:
     #     """
     #     parser = create_parser(lang)
     #     st = parser("A")
-    #     assert not st.errors and st.tag_name == "doc" and st.content == "A"
+    #     assert not st.errors and st.name == "doc" and st.content == "A"
     #     st = parser("E")
     #     assert st.errors and any(e.code == PARSER_STOPPED_BEFORE_END for e in st.errors)
     #
@@ -1152,7 +1371,7 @@ class TestSyntaxExtensions:
         st = parser('marke_8')
         assert not st.errors
         st = parser('t3vp ')
-        assert not st.errors
+        assert not st.errors, str(st.errors)
         st = parser('3tvp ')
         assert st.errors
         set_config_value('syntax_variant', 'strict')
@@ -1311,6 +1530,8 @@ class TestTreeOptimization:
 
 
 class TestHeuristics:
+    """Test of the "heuristic" "syntax_variant" (see DHParser.configuration)
+    which flexibly adjusts to the used EBNF-variant."""
     def setup(self):
         self.save = get_config_value('syntax_variant')
         set_config_value('syntax_variant', 'fixed')
@@ -1372,7 +1593,9 @@ class TestRuleOrder:
             """
         # set_config_value('compiled_EBNF_log', 'auto_reorder_test.log')
         parser = create_parser(reverse_order)
-        assert parser.B.__class__.__name__ != "Forward"
+        assert parser.B.__class__.__name__ == "Forward"
+        assert not parser.B.pname
+        assert parser.B.parser.pname == 'B'
 
     def test_recursive_root_detached_root(self):
         reverse_order = """
@@ -1483,6 +1706,67 @@ class TestRuleOrder:
         B = all_paths(compiler_obj)
         assert A == B, str(A) + str(B)
 
+
+class TestInclude:
+    number_ebnf = '''
+    @ disposable = DOT, EXP, FRAC, INT
+    NUMBER      = INT [FRAC] [EXP]
+    INT         = [NEG] ( /[1-9][0-9]+/ | /[0-9]/ )
+    NEG         = `-`
+    FRAC        = DOT /[0-9]+/
+    DOT         = `.`
+    EXP         = (`E`|`e`) [`+`|`-`] /[0-9]+/    
+    '''
+
+    arithmetic_ebnf = '''
+    @ whitespace  = vertical             
+    @ literalws   = right                
+    @ comment     = /#.*/                
+    @ ignorecase  = False     
+    @ reduction   = merge_treetops           
+    @ drop        = whitespace, strings    
+    
+    expression = term  { (add | sub) term}
+    term       = factor { (div | mul) factor}
+    factor     = [NEG] (number | group)
+    group      = "(" expression ")"
+    
+    add        = "+"
+    sub       = "-"
+    mul        = "*"
+    div        = "/"
+    
+    number     = NUMBER ~
+    
+    @ include = "number.ebnf"    
+    '''
+
+    def setup(self):
+        self.dirname = unique_name('test_ebnf_data')
+        os.mkdir(self.dirname)
+        self.number_path = os.path.join(self.dirname, "number.ebnf")
+        with open(self.number_path, 'w', encoding='utf-8') as f:
+            f.write(self.number_ebnf)
+
+    def teardown(self):
+        os.remove(self.number_path)
+        os.rmdir(self.dirname)
+
+    def test_include(self):
+        arithmetic_ebnf = self.arithmetic_ebnf.replace('number.ebnf', self.number_path)
+        # print(arithmetic_ebnf)
+        src, errors, ast = compile_ebnf(arithmetic_ebnf)
+        assert not errors, str(errors)
+        # for e in errors:
+        #     print(str(e), e.orig_doc)
+        # print(canonical_error_strings(errors))
+        parser = create_parser(arithmetic_ebnf)
+        # print(parser.python_src__)
+        tree = parser('2 - (3 * -4.145E+5)')
+        assert not tree.errors
+        # print(tree.as_sxpr())
+
+
 if __name__ == "__main__":
     from DHParser.testing import runner
     runner("", globals())
