@@ -33,7 +33,8 @@ from DHParser.toolkit import compile_python_object, re
 from DHParser.log import is_logging, log_ST, log_parsing_history, start_logging
 from DHParser.error import Error, is_error, add_source_locations, MANDATORY_CONTINUATION, \
     MALFORMED_ERROR_STRING, MANDATORY_CONTINUATION_AT_EOF, RESUME_NOTICE, PARSER_STOPPED_BEFORE_END, \
-    PARSER_NEVER_TOUCHES_DOCUMENT, CAPTURE_DROPPED_CONTENT_WARNING
+    PARSER_NEVER_TOUCHES_DOCUMENT, CAPTURE_DROPPED_CONTENT_WARNING, \
+    MANDATORY_CONTINUATION_AT_EOF_NON_ROOT
 from DHParser.parse import ParserError, Parser, Grammar, Forward, TKN, ZeroOrMore, RE, \
     RegExp, Lookbehind, NegativeLookahead, OneOrMore, Series, Alternative, \
     Interleave, CombinedParser, Text, EMPTY_NODE, Capture, Drop, Whitespace, \
@@ -43,7 +44,7 @@ from DHParser.ebnf import get_ebnf_grammar, get_ebnf_transformer, get_ebnf_compi
     parse_ebnf, DHPARSER_IMPORTS, compile_ebnf
 from DHParser.dsl import grammar_provider, create_parser
 from DHParser.preprocess import gen_neutral_srcmap_func
-from DHParser.syntaxtree import Node, parse_sxpr
+from DHParser.nodetree import Node, parse_sxpr
 from DHParser.stringview import StringView
 from DHParser.trace import set_tracer, trace_history, resume_notices_on
 
@@ -60,7 +61,7 @@ class TestWhitespace:
 
 class TestParserError:
     def test_parser_error_str(self):
-        pe = ParserError(Parser(), Node('TAG', 'test').with_pos(0),
+        pe = ParserError(Parser(), Node('TAG', 'test').with_pos(0), len('test'),
                          StringView('Beispiel'), None, first_throw=True)
         assert str(pe).find('Beispiel') >= 0 and str(pe).find('TAG') >= 0
 
@@ -147,7 +148,6 @@ class TestInfiLoopsAndRecursion:
             """
         snippet = "9 + 8 + 7 + 6 + 5 + 3 * 4"
         parser = grammar_provider(minilang)()
-        # print(raw_compileEBNF(minilang).result)
         assert parser
         syntax_tree = parser(snippet)
         if is_logging():
@@ -181,7 +181,6 @@ class TestInfiLoopsAndRecursion:
             Sum     = Expr { ('+' | '-') Expr }+
             Value   = /[0-9.]+/~ | '(' §Expr ')'
             """
-        # print(raw_compileEBNF(minilang).result)
         parser = grammar_provider(minilang)()
         snippet = "8 * 4"
         syntax_tree = parser(snippet)
@@ -341,7 +340,7 @@ class TestFlowControl:
         assert cst.error_flag, cst.as_sxpr()
 
         cst = parser(self.t2, parser['ws'], complete_match=False)
-        assert cst.did_match() and len(cst) == 0 and not cst.errors
+        assert cst.did_match() and cst.strlen() == 0 and not cst.errors
         cst = parser(self.t2, parser['word'], complete_match=False)
         assert cst.did_match() and cst.content == "All" and not cst.errors
         cst = parser(self.t2, parser['end'], complete_match=False)
@@ -381,7 +380,7 @@ class TestRegex:
         parser = compile_python_object(DHPARSER_IMPORTS + result, r'\w+Grammar$')()
         node = parser('abc+def', parser.regex)
         assert not node.error_flag
-        assert node.tag_name == "regex"
+        assert node.name == "regex"
         assert str(node) == 'abc+def'
 
     def test_multilineRegex_wo_Comments(self):
@@ -397,7 +396,7 @@ class TestRegex:
         parser = compile_python_object(DHPARSER_IMPORTS + result, r'\w+Grammar$')()
         node = parser('abc+def', parser.regex)
         assert not node.error_flag
-        assert node.tag_name == "regex"
+        assert node.name == "regex"
         assert str(node) == 'abc+def'
 
     def test_ignore_case(self):
@@ -413,7 +412,7 @@ class TestRegex:
         node, rest = parser.regex(StringView('Alpha'))
         assert node
         assert rest == ''
-        assert node.tag_name == "regex"
+        assert node.name == "regex"
         assert str(node) == 'Alpha'
 
         mlregex = r"""
@@ -467,12 +466,15 @@ class TestGrammar:
         # checks whether pos values in the parsing result and in the
         # history record have been initialized
         grammar = compile_python_object(DHPARSER_IMPORTS + self.pyparser, r'\w+Grammar$')()
+                                        # .format(dhparser_parentdir=repr('.')) + self.pyparser, r'\w+Grammar$')()
         grammar("no_file_name*")
         for record in grammar.history__:
             assert not record.node or record.node.pos >= 0
 
     def test_select_parsing(self):
-        grammar = compile_python_object(DHPARSER_IMPORTS + self.pyparser, r'\w+Grammar$')()
+        grammar = compile_python_object(
+            DHPARSER_IMPORTS + self.pyparser,
+            r'\w+Grammar$')()
         grammar("wort", "WORT")
         grammar("eine Zeile", "textzeile")
         grammar("kein Haupt", "haupt")
@@ -524,6 +526,27 @@ class TestGrammar:
         st = gr('-3')
         assert str(gr['S']) == "S = ~", str(gr['S'])
 
+    def test_match_and_fullmatch(self):
+        lang = r"""
+            word = /[A-Za-z]+/
+            number = /[0-9]+/
+        """
+        gr = create_parser(lang)
+        assert gr.match('word', 'hallo123') == "hallo"
+        assert gr.match('word', '') is None
+        assert gr.match('word', '123') is None
+        assert gr.match('word', 'hallo') == "hallo"
+
+        assert gr.fullmatch('word', 'hallo123') is None
+        assert gr.fullmatch('word', '') is None
+        assert gr.fullmatch('word', '123') is None
+        assert gr.fullmatch('word', 'hallo') == 'hallo'
+
+        assert gr.match('number', 'hallo123') is None
+        assert gr.match('number', '') is None
+        assert gr.match('number', '123') == "123"
+        assert gr.match('number', 'hallo') is None
+
 
 class TestSeries:
     def test_non_mandatory(self):
@@ -802,18 +825,22 @@ class TestPopRetrieve:
 
     @staticmethod
     def has_tag_name(node, name):
-        return node.tag_name == name # and not isinstance(node.parser, Retrieve)
+        return node.name == name # and not isinstance(node.parser, Retrieve)
 
     def test_capture_assertions(self):
         try:
-            _ = Grammar(Capture(Drop(Whitespace(r'\s*'))))
-            assert False, "GrammarError expected"
+            gr = Grammar(Capture(Drop(Whitespace(r'\s*'))))
+            assert any(ae.error.code == CAPTURE_DROPPED_CONTENT_WARNING
+                       for ae in gr.static_analysis_errors__), \
+                "Dropped content warning expected"
         except GrammarError as ge:
             assert ge.errors and ge.errors[0][-1].code == CAPTURE_DROPPED_CONTENT_WARNING, \
                 "Capture-dropped-content-Warning expected"
         try:
             _ = Grammar(Capture(Series(Text(' '), Drop(Whitespace(r'\s*')))))
-            assert False, "GrammarError expected"
+            assert any(ae.error.code == CAPTURE_DROPPED_CONTENT_WARNING
+                       for ae in gr.static_analysis_errors__), \
+                "Dropped content warning expected"
         except GrammarError as ge:
             assert ge.errors and ge.errors[0][-1].code == CAPTURE_DROPPED_CONTENT_WARNING, \
                 "Capture-dropped-content-Warning expected"
@@ -849,10 +876,6 @@ class TestPopRetrieve:
         assert not syntax_tree.error_flag, str(syntax_tree.errors_sorted)
 
     def test_optional_match(self):
-        # from DHParser.dsl import compileEBNF
-        # src = compileEBNF(self.mini_lang4)
-        # print(src)
-        # return
         test1 = '<info>Hey, you</info>'
         st = self.minilang_parser4(test1)
         assert not st.error_flag, str(st.errors_sorted)
@@ -865,7 +888,6 @@ class TestPopRetrieve:
         # start_logging('LOGS')
         st = self.minilang_parser4(test2)
         # log_parsing_history(self.minilang_parser4, "optional_match")
-        # print(st.as_sxpr())
         assert not st.error_flag
         test3 = '<info>Hey, <emph>you</></>'
         st = self.minilang_parser4(test3)
@@ -917,7 +939,6 @@ class TestPopRetrieve:
         # set_tracer(gr, trace_history)
         st = gr(case)
         # log_parsing_history(gr, 'test_cache_neutrality_2')
-        # print(st.as_sxpr())
         assert not st.errors
         assert str(st) == "AXX!"
 
@@ -935,8 +956,7 @@ class TestPopRetrieve:
         # set_tracer(gr, trace_history)
         st = gr(case)
         # log_parsing_history(gr, 'test_cache_neutrality_3')
-        # print(st.as_sxpr())
-        assert not st.errors, str(errors)
+        assert not st.errors, str(st.errors)
         case = 'AXXX!'
         st = gr(case)
         assert not st.errors
@@ -1010,8 +1030,6 @@ class TestPopRetrieve:
             value      = /\d+/~
             EOF        = !/./ [ :?defsign ]   # eat up captured defsigns
         """
-        # print(raw_compileEBNF(lang).result)
-        # set_config_value('compiled_EBNF_log', 'mylog.txt')
         parser = grammar_provider(lang)()
         st = parser("X := 1")
         assert not st.error_flag, str(st.errors)
@@ -1047,10 +1065,11 @@ class TestPopRetrieve:
         """
         parser = grammar_provider(lang_variant)()
         st = parser("X := 1")
-        assert not st.errors
+        assert not st.errors, str(st.errors)
         assert st.equals(st1)
         st = parser('')
-        assert "'EOF' expected" in str(st.errors), st.as_sxpr()
+        # for e in st.errors: print(e)
+        assert "'EOF = !/./ :?defsign' expected" in str(st.errors), st.as_sxpr()
 
 
 class TestWhitespaceHandling:
@@ -1224,7 +1243,7 @@ class TestReentryAfterError:
         content = 'ALPHA acb BETA bac GAMMA cab .'
         cst = gr(content)
         assert cst.error_flag
-        assert cst.content == content
+        assert cst.content == content, cst.content
         assert cst.pick('alpha').content.startswith('ALPHA')
 
     def test_no_resume_rules_partial_parsing(self):
@@ -1232,7 +1251,7 @@ class TestReentryAfterError:
         content = 'ALPHA acb'
         cst = gr(content, 'alpha')
         assert cst.error_flag
-        assert cst.content == content
+        assert cst.content == content, str(cst.content)
         assert cst.pick('alpha').content.startswith('ALPHA')
 
     def test_simple_resume_rule(self):
@@ -1241,7 +1260,7 @@ class TestReentryAfterError:
         content = 'ALPHA acb BETA bac GAMMA cab .'
         cst = gr(content)
         assert cst.error_flag
-        assert cst.content == content
+        assert cst.content == content, str(cst.content)
         assert cst.pick('alpha').content.startswith('ALPHA')
         # because of resuming, there should be only one error message
         assert len([err for err in cst.errors_sorted if err.code >= 1000]) == 1
@@ -1252,7 +1271,7 @@ class TestReentryAfterError:
         content = 'ALPHA acb BETA bac GAMMA cab .'
         cst = gr(content)
         assert cst.error_flag
-        assert cst.content == content
+        assert cst.content == content, cst.content
         # assert cst.pick('alpha').content.startswith('ALPHA')
 
     def test_several_reentry_points(self):
@@ -1261,7 +1280,7 @@ class TestReentryAfterError:
         content = 'ALPHA acb BETA bac GAMMA cab .'
         cst = gr(content)
         assert cst.error_flag
-        assert cst.content == content
+        assert cst.content == content, str(cst.content)
         assert cst.pick('alpha').content.startswith('ALPHA')
         # because of resuming, there should be only one error message
         assert len([err for err in cst.errors_sorted if err.code >= 1000]) == 1
@@ -1272,7 +1291,7 @@ class TestReentryAfterError:
         content = 'ALPHA acb GAMMA cab .'
         cst = gr(content)
         assert cst.error_flag
-        assert cst.content == content
+        assert cst.content == content, str(cst.content)
         assert cst.pick('alpha').content.startswith('ALPHA')
         # because of resuming, there should be only one error message
         assert len(cst.errors_sorted) == 1
@@ -1288,7 +1307,7 @@ class TestReentryAfterError:
         content = 'ALPHA abc BETA bad GAMMA cab .'
         cst = gr(content)
         assert cst.error_flag
-        assert cst.content == content
+        assert cst.content == content, str(cst.content)
         assert cst.pick('alpha').content.startswith('ALPHA')
         # because of resuming, there should be only one error message
         assert len([err for err in cst.errors_sorted if err.code >= 1000]) == 1
@@ -1440,46 +1459,46 @@ class TestMetaParser:
     mp.grammar = Grammar()  # override placeholder warning
     mp.pname = "named"
     mp.disposable = False
-    mp.tag_name = mp.pname
+    mp.node_name = mp.pname
 
     def test_return_value(self):
         nd = self.mp._return_value(Node('tagged', 'non-empty'))
-        assert nd.tag_name == 'named', nd.as_sxpr()
+        assert nd.name == 'named', nd.as_sxpr()
         assert len(nd.children) == 1
-        assert nd.children[0].tag_name == 'tagged'
+        assert nd.children[0].name == 'tagged'
         assert nd.children[0].result == "non-empty"
         nd = self.mp._return_value(Node('tagged', ''))
-        assert nd.tag_name == 'named', nd.as_sxpr()
+        assert nd.name == 'named', nd.as_sxpr()
         assert len(nd.children) == 1
-        assert nd.children[0].tag_name == 'tagged'
+        assert nd.children[0].name == 'tagged'
         assert not nd.children[0].result
         nd = self.mp._return_value(Node(':anonymous', 'content'))
-        assert nd.tag_name == 'named', nd.as_sxpr()
+        assert nd.name == 'named', nd.as_sxpr()
         assert not nd.children
         assert nd.result == 'content'
         nd = self.mp._return_value(Node(':anonymous', ''))
-        assert nd.tag_name == 'named', nd.as_sxpr()
+        assert nd.name == 'named', nd.as_sxpr()
         assert not nd.children
         assert not nd.content
         nd = self.mp._return_value(EMPTY_NODE)
-        assert nd.tag_name == 'named' and not nd.children, nd.as_sxpr()
+        assert nd.name == 'named' and not nd.children, nd.as_sxpr()
         self.mp.pname = ''
         self.mp.disposable = True
-        self.mp.tag_name = ':unnamed'
+        self.mp.node_name = ':unnamed'
         nd = self.mp._return_value(Node('tagged', 'non-empty'))
-        assert nd.tag_name == 'tagged', nd.as_sxpr()
+        assert nd.name == 'tagged', nd.as_sxpr()
         assert len(nd.children) == 0
         assert nd.content == 'non-empty'
         nd = self.mp._return_value(Node('tagged', ''))
-        assert nd.tag_name == 'tagged', nd.as_sxpr()
+        assert nd.name == 'tagged', nd.as_sxpr()
         assert len(nd.children) == 0
         assert not nd.content
         nd = self.mp._return_value(Node(':anonymous', 'content'))
-        assert nd.tag_name == ':anonymous', nd.as_sxpr()
+        assert nd.name == ':anonymous', nd.as_sxpr()
         assert not nd.children
         assert nd.result == 'content'
         nd = self.mp._return_value(Node('', ''))
-        assert nd.tag_name == '', nd.as_sxpr()
+        assert nd.name == '', nd.as_sxpr()
         assert not nd.children
         assert not nd.content
         assert self.mp._return_value(None) == EMPTY_NODE
@@ -1487,10 +1506,9 @@ class TestMetaParser:
 
     def test_return_values(self):
         self.mp.pname = "named"
-        self.mp.tag_name = self.mp.pname
+        self.mp.node_name = self.mp.pname
         rv = self.mp._return_values((Node('tag', 'content'), EMPTY_NODE))
-        # print(rv.as_sxpr())
-        assert rv[-1].tag_name != EMPTY_NODE.tag_name, rv[-1].tag_name
+        assert rv[-1].name != EMPTY_NODE.name, rv[-1].name
 
     def test_in_context(self):
         minilang = r"""
@@ -1566,8 +1584,10 @@ class TestStaticAnalysis:
     def test_cannot_capture_dropped_content(self):
         p = Capture(Drop(Whitespace(" ")))
         try:
-            _ = Grammar(p)
-            assert False, "GrammarError expected"
+            gr = Grammar(p)
+            assert any(ae.error.code == CAPTURE_DROPPED_CONTENT_WARNING
+                       for ae in gr.static_analysis_errors__), \
+                "Dropped content warning expected"
         except GrammarError as ge:
             assert ge.errors and ge.errors[0][-1].code == CAPTURE_DROPPED_CONTENT_WARNING, \
                 "Capture-dropped-content-Warning expected"
@@ -1587,8 +1607,6 @@ class TestStaticAnalysis:
         assert any(e.code == PARSER_NEVER_TOUCHES_DOCUMENT for e in errors)
         code, errors, ast = compile_ebnf(lang2, preserve_AST=True)
         assert any(e.code == PARSER_NEVER_TOUCHES_DOCUMENT for e in errors)
-        # for e in errors:
-        #     print(e)
 
 
 class TestMemoization:
@@ -1602,7 +1620,6 @@ class TestMemoization:
         EOF = /$/'''
         grammar = create_parser(words, 'words')
 
-        # print(grammar.python_src__)
         p1 = grammar.wordC.parsers[0]
         p2 = grammar.wordB.parsers[0]
         p3 = grammar.wordA.parsers[0]
@@ -1654,6 +1671,119 @@ class TestStringAlternative:
         assert longest_match(l, 'a', 2) == ''
 
 
+class TestErrorLocations:
+    def test_error_locations_bug(self):
+        miniXML = '''
+        @ disposable  = EOF
+        @ drop        = EOF, whitespace, strings
+        document = ~ element ~ §EOF
+        element  = STag §content ETag
+        STag     = '<' TagName §'>'
+        @ETag_skip = /[^<>]*/
+        ETag     = '</' § ::TagName '>'
+        TagName  = /\\w+/
+        content  = [CharData] { (element | COMMENT__) [CharData] }
+        CharData = /(?:(?!\\]\\]>)[^<&])+/
+        EOF      =  !/./
+        '''
+        testdoc = '''
+        <doc>
+            <title>Heading <wrong></title>
+        </doc>'''
+        parseXML = create_parser(miniXML)
+        _ = parseXML(testdoc)
+
+    def test_error_resumption(self):
+        miniXML = '''
+        @ whitespace  = /\s*/
+        @ disposable  = EOF
+        @ drop        = EOF, whitespace, strings
+
+        document = ~ element ~ §EOF
+        @element_resume = (:?TagName)
+        element  = STag §content ETag
+        @STag_skip = (/[^<>]*>/)
+        STag     = '<' TagName §'>'
+        @ETag_skip = (/[^<>]*/)
+        ETag     = '</' ::TagName §'>'
+        TagName  = /\w+/
+        content  = [CharData] { (element | COMMENT__) [CharData] }
+
+        CharData = /(?:(?!\]\]>)[^<&])+/
+        EOF      =  !/./        # no more characters ahead, end of file reached
+        '''
+        testdoc = '''
+        <doc>
+            <title>Heading <wrong></title>
+        </doc>'''
+        parseXML = create_parser(miniXML)
+        resume_notices_on(parseXML)
+        result = parseXML(testdoc)
+        assert len(result.errors) == 2
+
+    def test_error_resumption_2(self):
+        miniXML = '''
+        @ whitespace  = /\s*/
+        @ disposable  = EOF
+        @ drop        = EOF, whitespace, strings
+
+        document = ~ element ~ §EOF
+        @element_resume = /[^<>]*/
+        element  = STag §content ETag
+        @STag_skip = (/[^<>]*>/)
+        STag     = '<' TagName §'>'
+        @ETag_skip = (/[^<>]*/)
+        ETag     = '</' ::TagName §'>'
+        TagName  = /\w+/
+        content  = [CharData] { (element | COMMENT__) [CharData] }
+
+        CharData = /(?:(?!\]\]>)[^<&])+/
+        EOF      =  !/./        # no more characters ahead, end of file reached
+        '''
+        testdoc = '''
+        <doc>
+            <title>Heading <wrong></title>
+        </doc>'''
+        parseXML = create_parser(miniXML)
+        resume_notices_on(parseXML)
+        result = parseXML(testdoc)
+        # for e in result.errors:  print(e)
+        assert len(result.errors) == 2
+
+    def test_error_location(self):
+        grammar = r'''
+            @ string_error  = /\\/, 'Illegal escape sequence »{1}«'
+            @ string_error  = '', 'Illegal character "{1}" in string.'
+            @ string_resume = /("\s*)/
+            string          = `"` §characters `"` ~
+            characters      = { plain | escape }
+            plain           = /[^"\\]+/
+            escape          = /\[\/bnrt\\]/'''
+        json_string = create_parser(grammar, 'json_string')
+        tree = json_string('"al\\pha"')
+        assert len(tree.errors) == 1
+
+
+class TestStructurePreservationOnLookahead:
+    def test_structure_preservation_on_lookahead(self):
+        bib_grammar = r'''@literalws = right
+        @whitespace = horizontal
+        biliography = author ":" work { /\s*/ author ":" work } EOF 
+        author = name § { ~ name}+ &`:`
+        work = word § { ~ word} &(/\n/ | EOF)
+        name = word
+        word = /\w+/
+        EOF = !/./
+        '''
+        document = "Bertrand Russel: Principia Mathematica"
+        gr = create_parser(bib_grammar)
+        bib = gr(document)
+        assert bib['author'].content == "Bertrand Russel"
+        assert bib['work'].content == "Principia Mathematica"
+        author = gr('Bertrand Russell', 'author')
+        assert author.name == "author" and author.content == "Bertrand Russell"
+        assert author.errors[0].code == MANDATORY_CONTINUATION_AT_EOF_NON_ROOT
+
 
 if __name__ == "__main__":
     from DHParser.testing import runner
